from enum import Enum
from threading import Timer
import random
from ahc.Ahc import Topology, ComponentRegistry, ComponentModel, ConnectorTypes, Event, GenericMessageHeader, \
    GenericMessage, EventTypes
from ahc.Channels.Channels import Channel

registry = ComponentRegistry()
class ChandraMessageTypes(Enum):
    PRP = "propose"
    EST = "estimate"
    ACK = "acknowledge"
    NACK = "negacknowledge"
    DCD = "decide"
    IS_ALIVE = "iscoordinatoralive"

class ChandraPhases(Enum):
    IDLE = "idle"
    PRP_PHASE = "proposalphase"
    EST_PHASE = "estimatephase"
    ACK_PHASE = "ackphase"
    DCD_PHASE = "decisionphase"


class RepeatTimer(Timer):
    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


def selectCoordinator():
    cmp = get_coordinator()
    crdname = cmp.componentname
    crdinstancenumber = (cmp.componentinstancenumber + 1) % cmp.numberofNodes
    registry.get_component_by_key(crdname, crdinstancenumber).set_as_coordinator()
    cmp.set_as_proposer()

def start_next_round():
    print(f"No agreement. Next round coming up..")
    selectCoordinator()
    print(f"New coordinator is {get_coordinator().componentname}-{get_coordinator().componentinstancenumber}")
    cnodelist = list()
    for itemkey in registry.components:
        cmp = registry.components[itemkey]
        if cmp.componentname == "ChandraNode":
            cnodelist.append(cmp)
            #print(f"{cmp.componentname}-{cmp.componentinstancenumber}={type(cmp)}")
    for cmp in cnodelist:
        cmp.phase = ChandraPhases.IDLE
        cmp.faulty = random.randrange(100) > 95
        cmp.timer.cancel()
        cmp.timer = RepeatTimer(0.5, cmp.phase_handler)
        cmp.timer.start()
        cmp.start = True

def get_coordinator():
    cnodelist= list()
    for itemkey in registry.components:
        cmp = registry.components[itemkey]
        if cmp.componentname == "ChandraNode":
            cnodelist.append(cmp)

    for n in range(0,len(cnodelist)):
        if cnodelist[n].is_coordinator():
            return cnodelist[n]

class ChandraComponents(ComponentModel):
    def __init__(self, componentname, componentinstancenumber, numberofNodes):
        super().__init__(componentname, componentinstancenumber)
        self.phasehandlers = {ChandraPhases.IDLE: self.on_idle_phase, ChandraPhases.PRP_PHASE: self.on_proposal_phase,
                              ChandraPhases.EST_PHASE: self.on_estimate_phase,
                              ChandraPhases.ACK_PHASE: self.on_ACK_phase,
                              ChandraPhases.DCD_PHASE: self.on_decide_phase}
        self.numberofNodes = numberofNodes
        self.isCoordinator = False

        self.start = False
        self.phase = ChandraPhases.IDLE
        self.timer = RepeatTimer(0.5, self.phase_handler)
        self.count_ACK = 0
        self.count_proposal = 0

        self.value = random.randint(0, 3)
        self.timestamp = random.randint(0, 5)
        self.estimate = (self.value, self.timestamp)

        self.faulty = random.randrange(100) > 95 # Randomise faulty detection of coordinator



        print(f"Initial values for {self.componentname}-{self.componentinstancenumber} : "
              f"(Value,Timestamp)=({self.value},{self.timestamp})")

    def on_init(self, eventobj: Event):
        self.start = True
        self.timer.start()

    def on_message_from_bottom(self, eventobj: Event):
        message = eventobj.eventcontent
        msgtype = message.header.messagetype
        msgfrom = message.header.messagefrom

        messagePayload = message.payload
        value, timestamp = [int(n) for n in messagePayload.split("-")]

        if msgtype is ChandraMessageTypes.PRP:
            print(f"{msgfrom.componentname}-{msgfrom.componentinstancenumber} sent an proposal message to "
                  f"{self.componentname}-{self.componentinstancenumber} = ({value},{timestamp})")
            self.count_proposal = self.count_proposal + 1
            if timestamp > self.estimate[1]:
                self.estimate = (value, timestamp)

        elif msgtype is ChandraMessageTypes.EST:
            self.estimate = (value, timestamp)
            print(f"{msgfrom.componentname}-{msgfrom.componentinstancenumber} sent an estimate message to "
                  f"{self.componentname}-{self.componentinstancenumber} = ({value},{timestamp})")


        elif msgtype is ChandraMessageTypes.ACK:
            print(f"{msgfrom.componentname}-{msgfrom.componentinstancenumber} sent an ACK message to "
                  f"{self.componentname}-{self.componentinstancenumber}")
            self.count_ACK = self.count_ACK + 1

        elif msgtype is ChandraMessageTypes.NACK:
            print(f"{msgfrom.componentname}-{msgfrom.componentinstancenumber} sent an NACK message to "
                  f"{self.componentname}-{self.componentinstancenumber}")
            self.phase = ChandraPhases.IDLE
            start_next_round()

        # change the new value with the decision taken by coordinator
        elif msgtype is ChandraMessageTypes.DCD:
            self.value = value
            self.timestamp = timestamp
            print(f"Decision message is arrived. The new values and time stamps of the component "
                  f"{self.componentname}-{self.componentinstancenumber} = ({value},{timestamp})")

        elif msgtype is ChandraMessageTypes.FDI:
            # FailureDetector message
            pass

    def phase_handler(self):
        # print(f"{self.componentname}.{self.componentinstancenumber} previous phase is {self.phase}")
        self.phasehandlers[self.phase]()
        # print(f"{self.componentname}.{self.componentinstancenumber} next phase is {self.phase}")

    def on_idle_phase(self):
        if self.start:
            self.phase = ChandraPhases.PRP_PHASE
        else:
            self.timer.cancel()

    def on_proposal_phase(self):
        self.phase = ChandraPhases.EST_PHASE
        if not self.is_coordinator():
            self.send_propose_message()
        else:
            pass

    def on_estimate_phase(self):
        self.phase = ChandraPhases.ACK_PHASE
        if self.is_coordinator():
            if self.count_proposal >= (len(registry.components) + 1) / 2:
                print(f"The estimate of coordinator is: {self.estimate}")
                self.send_estimate_message()
            else:
                start_next_round()
        else:
            pass

    def on_ACK_phase(self):
        self.phase = ChandraPhases.DCD_PHASE
        if not self.is_coordinator():
            if not self.faulty:
                self.send_ACK()
            else:
                self.send_NACK()
        else:
            pass

    def on_decide_phase(self):
        self.phase = ChandraPhases.IDLE
        self.start = False
        if self.is_coordinator():
            if self.count_ACK >= (len(registry.components) + 1) / 2:
                self.send_decide_message()
            else:
                start_next_round()
        else:
            pass

    def send_propose_message(self):
        coordinator = get_coordinator()
        if self is not coordinator:
            cHeader = GenericMessageHeader(ChandraMessageTypes.PRP, self, coordinator)
            cPayload = f"{self.value}-{self.timestamp}"
            cEventContent = GenericMessage(cHeader, cPayload)
            cEvent = Event(self, EventTypes.MFRT, cEventContent)
            self.send_down(cEvent)
            self.send_self(cEvent)
        else:
            pass

    def send_estimate_message(self):
        coordinator = get_coordinator()
        if self is coordinator:
            cHeader = GenericMessageHeader(ChandraMessageTypes.EST, self, self.value, self.timestamp)
            cPayload = f"{self.estimate[0]}-{self.estimate[1]}"
            cEventContent = GenericMessage(cHeader, cPayload)
            cEvent = Event(self, EventTypes.MFRT, cEventContent)
            self.send_down(cEvent)
        else:
            pass

    def send_ACK(self):
        coordinator = get_coordinator()
        if self is not coordinator:
            cHeader = GenericMessageHeader(ChandraMessageTypes.ACK, self, coordinator)
            cPayload = f"{self.estimate[0]}-{self.estimate[1]}"
            cEventContent = GenericMessage(cHeader, cPayload)
            cEvent = Event(self, EventTypes.MFRT, cEventContent)
            self.send_down(cEvent)
        else:
            pass

    def send_NACK(self):
        coordinator = get_coordinator()
        if self is not coordinator:
            cHeader = GenericMessageHeader(ChandraMessageTypes.NACK, self, coordinator)
            cPayload = f"{self.estimate[0]}-{self.estimate[1]}"
            cEventContent = GenericMessage(cHeader, cPayload)
            cEvent = Event(self, EventTypes.MFRT, cEventContent)
            self.send_down(cEvent)
        else:
            pass

    def send_decide_message(self):
        coordinator = get_coordinator()
        if self is coordinator:
            cHeader = GenericMessageHeader(ChandraMessageTypes.DCD, self, None)
            cPayload = f"{self.estimate[0]}-{self.estimate[1]}"
            cEventContent = GenericMessage(cHeader, cPayload)
            cEvent = Event(self, EventTypes.MFRT, cEventContent)
            self.send_down(cEvent)
            self.value = self.estimate[0]
            self.timestamp = self.estimate[1]
            print(f"Agreement done. Change your values with decision: {cPayload}")
        else:
            pass

    def on_message_from_top(self, eventobj: Event):
        pass

    def set_as_coordinator(self):
        self.isCoordinator = True

    def set_as_proposer(self):
        self.isCoordinator = False
        self.count_ACK = 0
        self.count_proposal = 0

    def is_coordinator(self):
        return self.isCoordinator

class ChandraChannel(Channel):
  # Overwrite onDeliverToComponent if you want to do something in the last pipeline stage
  # onDeliver will deliver the message from the channel to the receiver component using messagefromchannel event
  def on_deliver_to_component(self, eventobj: Event):
    coordinator = get_coordinator()
    crdname = coordinator.componentinstancenumber
    callername = eventobj.eventsource.componentinstancenumber
    for item in self.connectors:
      callees = self.connectors[item]
      for callee in callees:
        calleename = callee.componentinstancenumber
        # print(f"I am connected to {calleename}. Will check if I have to distribute it to {item}")
        if calleename == callername or (calleename!=crdname and callername !=crdname):
          pass
        else:
          myevent = Event(eventobj.eventsource, EventTypes.MFRB, eventobj.eventcontent, self.componentinstancenumber)
          callee.trigger_event(myevent)

class ChandraConsensusNode(ComponentModel):

    def __init__(self, componentname, componentid, numberofNodes=10):
        self.channel = ChandraChannel("ChandraChannel", 0)
        for n in range(0, numberofNodes):
            tComp = ChandraComponents("ChandraNode", n, numberofNodes)
            tComp.connect_me_to_channel(ConnectorTypes.DOWN, self.channel)
            registry.add_component(tComp)

        registry.get_component_by_key("ChandraNode", 0).set_as_coordinator()
        super().__init__(componentname, componentid)

def main():
    topo = Topology()
    topo.construct_single_node(ChandraConsensusNode, 0)
    topo.start()
    while True: pass


if __name__ == '__main__':
    main()



from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, ComponentRegistry, Lock, \
    Thread, Topology
from enum import Enum


class OLSREventTypes(Enum):
    HELLO = "HELLO_MESSAGE"
    TC = "TC_MESSAGE"



class OLSRMessageTypes(Enum):
    GENERATE_HELLO = 'GENERATE_HELLO'
    GENERATE_TC = 'GENERATE_TC'
    HELLO = 'HELLO'
    TC = "TC"
    UPDATE_TABLE="UPDATE_TABLE"


class OLSRSimplifiedRouter(ComponentModel):
    def __init__(self, componentname, componentid):
        super(OLSRSimplifiedRouter, self).__init__(componentname, componentid)
        self.eventhandlers[OLSREventTypes.HELLO] = self.on_hello
        self.eventhandlers[OLSREventTypes.TC] = self.on_tc


        self.duplicateSet = []
        self.linkSet = []
        self.twoHopNeighborSet = []
        self.topologySet = []
        self.neighborSet = []
        self.routingTable=[]
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber)

        self.send_self(Event(self, EventTypes.INIT, None))
    def on_init(self, eventobj: Event):
        super(OLSRSimplifiedRouter, self).on_init(eventobj)
        self.neighborUpdate()

    def neighborUpdate(self):

        for node in self.neighbors:
            self.linkSet.append(node)
        print(self.linkSet)

    def on_message_from_peer(self, eventobj: Event):
        payload = eventobj.eventcontent

        if payload == OLSRMessageTypes.UPDATE_TABLE:
            print('update table')
            self.update_routing_table()
        if payload == OLSRMessageTypes.GENERATE_HELLO:
            print('in hello')
            payload2 = []
            print(self.linkSet)
            for item in self.linkSet:
                print('first for')
                payload2.append(item)
            for node in self.neighborSet:
                print('second for')
                header= OLSRMessageHeader(OLSRSimplifiedRouter.__name__+"-"+self.componentinstancenumber,OLSRMessageTypes.HELLO,
                                           OLSRSimplifiedRouter.__name__+"-"+self.componentinstancenumber,OLSRSimplifiedRouter.__name__ + "-" + str(node),
                                           str(self.componentinstancenumber) + "-" + str(node))
                eventContent = GenericMessage(header, payload2)
                print("sent hello")
                self.send_down(Event(self, EventTypes.MFRT, eventContent))
        if payload == OLSRMessageTypes.GENERATE_TC:
            payload2=[]
            for node in self.linkSet:
                payload2.append(node)
            for neighbor in self.neighborSet:
                header = OLSRMessageHeader(OLSRSimplifiedRouter.__name__ + "-" + self.componentinstancenumber,
                                            OLSRMessageTypes.TC,
                                            OLSRSimplifiedRouter.__name__ + "-" + self.componentinstancenumber,
                                            OLSRSimplifiedRouter.__name__ + "-" + str(node),
                                            str(self.componentinstancenumber) + "-" + str(node))
                eventContent = GenericMessage(header, payload2)
                self.send_down(Event(self, EventTypes.MFRT, eventContent))

    def on_message_from_bottom(self, eventobj: Event):
        print('on message from bottom')
        message = Event.eventcontent
        messageHeader = message.header
        messageContent = message.payload

        if messageContent is None:
            print(f"{self.unique_name()} Message dropped from {messageHeader.messagefrom} is empty.")
            return
        else:
            self.should_message_be_processed(messageContent,messageHeader)
            self.should_message_be_forwarded(messageContent, messageHeader)

    def should_message_be_processed(self, messageHeader,messageContent):
        if any(duplicate.get('D_addr') == messageHeader.originatorAddress and
                duplicate.get('D_seq_num') == messageHeader.messageSequenceNumber for duplicate in self.duplicateSet):
             print(f"{self.unique_name()} Message has not been processed coming from {messageHeader.messagefrom},it has already been processed.")
             return
        if messageHeader.messagetype in OLSRMessageTypes:
            print(
                f"{self.unique_name()} Message will be processed coming from {messageHeader.messagefrom}.")
            self.process_message(messageHeader,messageContent)

    def process_message(self,message_header,message_content):
        msg=GenericMessage(message_header,message_content)
        if message_header.messagetype==OLSRMessageTypes.HELLO: self.send_self(Event(self,OLSREventTypes.HELLO,msg))
        if message_header.messagetype==OLSRMessageTypes.TC: self.send_self(Event(self,OLSREventTypes.TC,msg))
    def forward_message(self,message_header,message_content):
        print('forward message')
        for neighbor in self.neighborSet:
            header = OLSRMessageHeader(message_header.originatoraddress,
                                       OLSRMessageTypes.TC,
                                       OLSRSimplifiedRouter.__name__ + "-" + self.componentinstancenumber,
                                       OLSRSimplifiedRouter.__name__ + "-" + str(neighbor),
                                       str(self.componentinstancenumber) + "-" + str(neighbor))
            eventContent = GenericMessage(header, message_content)
            self.send_down(Event(self, EventTypes.MFRT, eventContent))
    def update_routing_table(self):
        print('update routing table')
        #neighbor tabledakiler
        for node in self.linkSet:
            self.routingTable.append({'R_dest_addr':node,'R_next_addr':node,
                                      'R_dist':1,'R_iface_addr':self.componentinstancenumber})
        for node in self.twoHopNeighborSet:
            index = next(
                i for i, item in enumerate(self.routingTable) if item["R_dest_addr"] == node.N_neighbor_main_addr)
            R_next_addr=self.routingTable[index]['R_next_addr']
            R_iface=self.routingTable[index]['R_iface_addr']
            self.routingTable.append({'R_dest_addr':node.N_neighbor_main_addr,'R_next_addr':R_next_addr,
                                      'R_dist':2,'R_iface_addr':R_iface})
        for entry in self.topologySet:
            if not any(item.get('R_dest_addr')==entry.T_dest_addr for item in self.routingTable) and any(item.get('R_dest_addr')==entry.T_last_addr for
                                                                                                         item in self.routingTable):
                index = next(
                    i for i, item in enumerate(self.routingTable) if item["R_dest_addr"] == entry.T_last_addr)
                h=self.routingTable[index]['R_dist']
                if h>=2:
                    r_next=self.routingTable[index]['R_next_addr']
                    r_face=self.routingTable[index]['R_iface_addr']
                    self.routingTable.append({'R_dest_addr':entry.T_dest_addr,'R_next_addr':r_next,'R_dist':h+1,'R_iface_addr':r_face})
    def on_tc(self, eventobj: Event):
        print('on tc')
        message=eventobj.eventcontent
        hdr=message.header
        payload=message.payload
        if hdr.messagefrom not in self.neighborSet: return
        elif any(item.get('T_last_addr')==hdr.originatoraddress for item in self.topologySet): return
        else:
            for address in payload:
                if any(item.get('T_dest_addr')==address and item.get('T_last_addr')==hdr.originatoraddress for item in self.topologySet):return
                else:
                    self.topologySet.append({'T_dest_addr':address,'T_last_addr':hdr.originatoraddress})
    def on_hello(self, eventobj: Event):
        print('on hello')
        message=eventobj.eventcontent
        payload=message.payload
        hdr=message.header
        if hdr.originatoraddress is self.componentinstancenumber: return
        if hdr.originatoraddress not in self.linkSet:
            self.linkSet.append(hdr.messagefrom)
            self.neighborSet.append(hdr.messagefrom)
        for item in payload:
            if item not in self.neighborSet:
                if not any(element.get('N_neighbor_main_addr')==hdr.originatoraddress and element.get('N_2hop_addr')==item for element in self.twoHopNeighborSet):
                    self.twoHopNeighborSet.append({'N_neighbor_main_addr':hdr.originatoraddress,'N_2hop_addr':item})

    def should_message_be_forwarded(self, messageHeader,messageContent):
        if any(duplicate.get('D_addr') == messageHeader.originatorAddress and
            duplicate.get('D_seq_num') == messageHeader.messageSequenceNumber and
            duplicate.get('D_iface_list')==OLSRSimplifiedRouter.__name__+"-"+self.componentinstancenumber for duplicate in self.duplicateSet):
            print(f"{self.unique_name()} Message will not be forwarded coming from {messageHeader.messagefrom},it has already been considered for forwarding.")
            return

        if messageHeader.messagetype==OLSRMessageTypes.TC:
            self.forward_message(messageHeader,messageContent)
        else:
            self.default_forwarding(messageHeader,messageContent)

    def default_forwarding(self,messageHeader,messageContent):
        if messageHeader.messageto in self.neighborSet:
            msg=GenericMessage(messageHeader,messageContent)
            self.send_down(Event(self,EventTypes.MFRT,msg))
        else:
            return


class OLSRMessageHeader(GenericMessageHeader):
    def __init__(self,  originatoraddress,
                 messagetype, messagefrom, messageto,interface_id ):
        super(OLSRMessageHeader, self).__init__(messagetype, messagefrom, messageto,interfaceid=interface_id)
        self.originatorAddress = originatoraddress





from ahc.Ahc import ComponentModel, Event, ConnectorTypes, EventTypes, Topology
from ahc.Routing.DBR2P.AlgorithmDBR2PComponent import *
from ahc.Routing.DBR2P.Messages import *
from ahc.Routing.DBR2P.ApplicationComponent import *


# Encapsulator for the DBR2P Node
class DBR2PNode(ComponentModel):
    def __init__(self, componentname, componentid):
        super(DBR2PNode, self).__init__(componentname, componentid)
        self.ApplicationComponent = ApplicationComponent(ApplicationComponent.__name__, componentid)
        self.DBR2PAlgorithmComponent = RoutingDBR2PComponent(RoutingDBR2PComponent.__name__, componentid)

        self.inactive = False

        self.ApplicationComponent.connect_me_to_component(ConnectorTypes.DOWN, self.DBR2PAlgorithmComponent)
        self.DBR2PAlgorithmComponent.connect_me_to_component(ConnectorTypes.UP, self.ApplicationComponent)

        self.connect_me_to_component(ConnectorTypes.UP, self.DBR2PAlgorithmComponent)
        self.DBR2PAlgorithmComponent.connect_me_to_component(ConnectorTypes.DOWN, self)

        self.connect_me_to_component(ConnectorTypes.PEER, self.ApplicationComponent)

    def on_message_from_bottom(self, eventobj: Event):
        #print(f"{EventTypes.MFRB} {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} & {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")
        if not self.inactive:
            self.send_up(eventobj)  # send incoming messages to upper components
        else:
            self.time_to_wait -= (time.time() - self.time_wait_start)
            if self.time_to_wait <= 0:
                self.inactive = False

    def on_message_from_top(self, eventobj: Event):
        #print(f"{EventTypes.MFRT} {self.componentname}.{self.componentinstancenumber} => {eventobj.eventcontent.header.messagetype} & {eventobj.eventcontent.header.messagefrom} $ {eventobj.eventcontent.header.messageto}")

        if not self.inactive:
            self.send_down(eventobj)  # send incoming messages from upper components to a channel
        else:
            self.time_to_wait -= (time.time() - self.time_wait_start)
            if self.time_to_wait <= 0:
                self.inactive = False

    def inactive_for_time(self, time_to_wait):
        self.inactive = True
        self.time_to_wait = time_to_wait
        self.time_wait_start = time.time()

    def send_data_to(self, destination):
        if not self.inactive:
            header = DBR2PMessageHeader(DBR2PMesageType.DT, self.componentinstancenumber, destination)
            payload = DBR2PMessagePayload([])
            evt = Event(self, EventTypes.MFRP, GenericMessage(header, payload))
            self.send_peer(evt)
        else:
            self.time_to_wait -= (time.time() - self.time_wait_start)
            if self.time_to_wait <= 0:
                self.inactive = False

    def send_down(self, event: Event):
        try:
            for p in self.connectors[ConnectorTypes.DOWN]:
                event.eventcontent.header.interfaceid = p.componentinstancenumber
                #print(
                #    f"{EventTypes.MFRT} {self.componentname}.{self.componentinstancenumber} => {event.eventcontent.header.interfaceid} & {event.eventcontent.header.messagefrom} $ {event.eventcontent.header.messageto}")

                header = DBR2PMessageHeader(event.eventcontent.header.messagetype,
                                            event.eventcontent.header.messagefrom,
                                            event.eventcontent.header.messageto,
                                            event.eventcontent.header.nexthop,
                                            event.eventcontent.header.interfaceid,
                                            event.eventcontent.header.sequencenumber)
                messagepayload = copy.deepcopy(event.eventcontent.payload.messagepayload)
                payload = DBR2PMessagePayload(messagepayload, time_now=event.eventcontent.payload.time)
                evt = Event(self, EventTypes.MFRT, GenericMessage(header, payload))
                p.trigger_event(evt)
        except Exception as e:
            print(e)
            pass



from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes
from ahc.Routing.GSR.Constants import TERMINATE_MESSAGE_TYPE, \
    INFO_MESSAGE_TYPE, \
    ROUTING_COMPLETED_MESSAGE_TYPE, \
    GSR_COORDINATOR_NAME, \
    GSR_ROUTER_NAME, \
    GSR_APPLICATION_NAME, \
    ENABLE_NETWORK_LEVEL_LOGGING


class GSRCoordinator(ComponentModel):
    def __init__(self, component_name, component_id):
        super(GSRCoordinator, self).__init__(component_name, component_id)
        self.routing_table = {}

    def on_message_from_top(self, eventobj: Event):
        event_header = eventobj.eventcontent.header
        sender = event_header.messagefrom.split("-")[0]
        message_to = event_header.messageto.split("-")[0]
        message_type = event_header.messagetype
        message = eventobj.eventcontent.payload
        if ENABLE_NETWORK_LEVEL_LOGGING:
            print(f"Coordinator {self.componentinstancenumber} receives message from top")
        if message_to == GSR_COORDINATOR_NAME:
            if sender == GSR_APPLICATION_NAME and message_type == TERMINATE_MESSAGE_TYPE:
                self.terminate_routing()
            elif sender == GSR_APPLICATION_NAME and message_type == INFO_MESSAGE_TYPE:
                self.route_message(message)

    def on_message_from_peer(self, eventobj: Event):
        event_header = eventobj.eventcontent.header
        sender = event_header.messagefrom.split("-")[0]
        message_to = event_header.messageto.split("-")[0]
        message = eventobj.eventcontent.payload
        if message_to == GSR_COORDINATOR_NAME:
            if sender == GSR_ROUTER_NAME:
                self.routing_table = message["routing_table"]
                if ENABLE_NETWORK_LEVEL_LOGGING:
                    print(f"Coordinator {self.componentinstancenumber} has received Routing Table")
                if self.componentinstancenumber == 0:
                    message_header = GenericMessageHeader(
                        ROUTING_COMPLETED_MESSAGE_TYPE,
                        GSR_COORDINATOR_NAME + "-0",
                        GSR_APPLICATION_NAME + "-0")
                    message = GenericMessage(message_header, {"n_nodes": len(self.routing_table)})
                    kickstarter = Event(self, EventTypes.MFRB, message)
                    self.send_up(kickstarter)

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        event_header = eventobj.eventcontent.header
        sender = event_header.messagefrom.split("-")[0]
        message_to = event_header.messageto.split("-")[0]
        message_type = event_header.messagetype
        message = eventobj.eventcontent.payload
        if message_to == GSR_COORDINATOR_NAME:
            if sender == GSR_COORDINATOR_NAME and message_type == INFO_MESSAGE_TYPE:
                dest = message["dest"]
                src = message["src"]
                if ENABLE_NETWORK_LEVEL_LOGGING:
                    print(f"Coordinator {self.componentinstancenumber} has received info message from {src} to {dest}")

                if dest == self.componentinstancenumber:
                    message_header = GenericMessageHeader(
                        message_type,
                        GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber),
                        GSR_APPLICATION_NAME + "-" + str(self.componentinstancenumber))
                    message_to_application = GenericMessage(message_header, message)
                    event = Event(self, EventTypes.MFRB, message_to_application)
                    self.send_up(event)
                    # send to app layer
                    pass
                else:
                    self.route_message(message)

    def terminate_routing(self):
        message_from = GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber)
        message_to = GSR_ROUTER_NAME + "-" + str(self.componentinstancenumber)
        message_header = GenericMessageHeader(TERMINATE_MESSAGE_TYPE,
                                              message_from,
                                              message_to)
        message = GenericMessage(message_header, "")
        event = Event(self, EventTypes.MFRP, message)
        self.send_peer(event)
        if ENABLE_NETWORK_LEVEL_LOGGING:
            print("SENDING [" + message_from + " -> " + message_to + "]: TERMINATE")

    def route_message(self, message_to_route):
        if len(self.routing_table) > 0:
            message_from = GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber)
            dest = message_to_route["dest"]
            neighbor_id = self.routing_table[dest]
            message_to = GSR_COORDINATOR_NAME + "-" + str(neighbor_id)
            interface_id = str(self.componentinstancenumber) + "-" + str(neighbor_id)
            message_header = GenericMessageHeader(
                INFO_MESSAGE_TYPE,
                message_from,
                message_to,
                interfaceid=interface_id
            )
            message = GenericMessage(message_header, message_to_route)

            event = Event(self, EventTypes.MFRT, message)
            self.send_down(event)
            if ENABLE_NETWORK_LEVEL_LOGGING:
                print("ROUTING [" + message_from + " -> " + message_to + "]: " + str(message.payload))



from ahc.Ahc import ComponentModel, Event, EventTypes, GenericMessage
from ahc.Routing.DBR2P.Messages import *

import copy

class ApplicationComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(ApplicationComponent, self).__init__(componentname, componentid)
        self.routes = {}
        self.message_arrival_times = {}
        self.message_route_discovery_times = {}
        self.message_route_discovery_from_backup_times = {}
    def on_message_from_bottom(self, eventobj: Event):
        if eventobj.eventcontent.header.messagetype == DBR2PMesageType.ERROR:
            try:
                self.routes.pop(eventobj.eventcontent.header.messageto)
            except:
                pass
            print("{ERROR:{" + f"{eventobj.eventcontent.header.messagefrom}:{eventobj.eventcontent.header.messageto}"+"}")
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.RDReply:
            delta = time.time() - eventobj.eventcontent.payload.time
            dest = eventobj.eventcontent.header.messagefrom
            try:
                self.message_route_discovery_times[dest].append(delta)
            except KeyError:
                self.message_route_discovery_times[dest] = [delta]
            self.routes[dest] = eventobj.eventcontent.payload.messagepayload
            print("{"+f"RouteDiscovered:{time.time() - eventobj.eventcontent.payload.time},"+"{"+f"Route:{eventobj.eventcontent.payload.messagepayload}"+"}"+"}")
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.DT:
            delta = time.time() - eventobj.eventcontent.payload.time
            try:
                self.message_arrival_times[eventobj.eventcontent.header.messagefrom].append(delta)
            except KeyError:
                self.message_arrival_times[eventobj.eventcontent.header.messagefrom] = [delta]
            #print("{"+f"MessageReceived:{time.time() -eventobj.eventcontent.payload.time}"+"}")
        elif eventobj.eventcontent.header.messagetype == DBR2PMesageType.LRM:
            delta = time.time() - eventobj.eventcontent.payload.time
            dest = eventobj.eventcontent.payload.messagepayload[-1]
            try:
                self.message_route_discovery_from_backup_times[dest].append(delta)
            except KeyError:
                self.message_route_discovery_from_backup_times[dest] = [delta]
            self.routes[dest] = eventobj.eventcontent.payload.messagepayload
            print("{" + f"RouteDiscoveredFromBackup:{time.time() - eventobj.eventcontent.payload.time}"+"}")
        else:
            print(f"{eventobj.eventcontent.header.messagetype}")

    def on_message_from_peer(self, eventobj: Event):
        if eventobj.eventcontent.header.messagetype == DBR2PMesageType.DT:
            try:
                route = copy.deepcopy(self.routes[eventobj.eventcontent.header.messageto])
                header = DBR2PMessageHeader(DBR2PMesageType.DT, self.componentinstancenumber, eventobj.eventcontent.header.messageto)
                payload = DBR2PMessagePayload(route)
                message = GenericMessage(header, payload)
                message.payload.time = time.time()
                self.send_down(Event(self, EventTypes.MFRT, message))
            except KeyError:
                header = DBR2PMessageHeader(DBR2PMesageType.RDRequest, self.componentinstancenumber,
                                            eventobj.eventcontent.header.messageto)
                payload = DBR2PMessagePayload([])
                message = GenericMessage(header, payload)
                message.payload.time = time.time()
                self.send_down(Event(self, EventTypes.MFRT, message))
        else:
            print(f"{eventobj.eventcontent.header.messagetype}")



import matplotlib.pyplot as plt
import numpy as np
import pickle
import os

results = {}

for node_numbers in (5,10,15,20,25,30):
    result_dir = os.getcwd()+"/Routing/ARA/Results/"+str(node_numbers)
    files = os.listdir(result_dir)
    results[node_numbers] = {"Hops": [], "Time": []}
    print(f"{node_numbers} has {len(files)} results")
    total_file_count = len(files)
    for file in files:
        _ , total_time, total_hops = pickle.load(open(result_dir+"/"+file, "rb"))
        results[node_numbers]["Hops"].append(total_hops)
        results[node_numbers]["Time"].append(total_time)



nodes = ("5", "10", "15","20","25","30")
x_pos = np.arange(len(nodes)*2)

hop_mean = [np.mean(results[node_count]["Hops"]) for node_count in (5,10,15,20,25,30)]
hop_std = [np.std(results[node_count]["Hops"]) for node_count in (5,10,15,20,25,30)]
time_mean = [np.mean(results[node_count]["Time"]) for node_count in (5,10,15,20,25,30)]
time_std = [np.std(results[node_count]["Time"]) for node_count in (5,10,15,20,25,30)]

plots = ["Hops", "Time"]
plot = plots[1]

colors=["blue", "green"]
fig, ax = plt.subplots()
if plot == "Hops":
    ax.bar((1,2,3,4,5,6), hop_mean, yerr=hop_std, align='center', alpha=0.5, ecolor='black', capsize=10, color=colors[0], label="Message")
else:
    ax.bar((1,2,3,4,5,6), time_mean,  align='center', alpha=0.5, ecolor='black', capsize=10, color=colors[1], label="Time")

ax.set_xlabel("Node Count")
ax.set_xticks((1, 2, 3, 4,5,6))
ax.set_xticklabels(nodes)
if plot == "Hops":
    ax.set_ylabel('Total Hop Count')

    ax.set_title("Total Hop Count vs Node Count for ARA Route Discovery")
else:
    ax.set_ylabel('Elapsed Time (sec)')

    ax.set_title("Time vs Node Count for for ARA Route Discovery")

# ax.yaxis.grid(True)

# Save the figure and show
plt.tight_layout()
# plt.savefig('bar_plot_'+plot+'.png')
# plt.legend(loc=2)
plt.show()






import os
import sys
import time
import random
from enum import Enum

sys.path.insert(0, os.getcwd())

import networkx as nx
import matplotlib.pyplot as plt
import threading

from ahc.Ahc import ComponentModel, Event, ConnectorTypes, Topology
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Channels.Channels import P2PFIFOPerfectChannel
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
#from NetworkLayers.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer
from ahc.Routing.AODV.AODVNetworkLayerComponent import AODVNetworkLayerComponent
from ahc.Routing.AODV.AODVUtils import AODVMessageTypes, AODVMessageHeader

registry = ComponentRegistry()

class AODVApplicationLayerEventType(Enum):
      PROPOSE = "PROPOSE"

class AODVApplicationLayerComponent(ComponentModel):
    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

    def on_message_from_bottom(self, eventobj: Event):
        self.lock.acquire()
        
        msg = eventobj.eventcontent
        hdr = msg.header            
        #print(f"On MSFRB Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        if hdr.messagetype == AODVMessageTypes.RREP:
            self.send_self(Event(self,AODVApplicationLayerEventType.PROPOSE,eventobj.eventcontent))
        self.lock.release()

    def on_propose(self,eventobj:Event):
        self.lock.acquire()
        applmsg = eventobj.eventcontent
        applhdr = applmsg.header
        
        if applhdr.messagetype == AODVMessageTypes.RREP:
            print(f"Actual packet is ready to go from {self.componentname}.{self.componentinstancenumber} to {applhdr.messagefrom}")

            hdr = AODVMessageHeader(AODVMessageTypes.PROPOSE, self.componentinstancenumber,
                        applhdr.messagefrom,0,float('inf'),self.componentinstancenumber)
            data = self.MessageQueue[hdr.messageto]
            payload = GenericMessagePayload(data)
            message = GenericMessage(hdr, payload)
            #self.send_down(Event(self,EventTypes.MFRT,message))
        else:
            print(f"Wrong invocation of on_propose callback on {self.componentname}.{self.componentinstancenumber}")

        self.lock.release()

    #On-demand behavior sustained with this.
    def sendPackageToNode(self, destNodeID,msgFromUser):
        self.lock.acquire()
        #print(f"On sendPackageToNode {self.componentname}.{self.componentinstancenumber}")
        if self.componentinstancenumber == destNodeID:
            print(f"Node tries to sent itself so aborting")
            return
        hdr = AODVMessageHeader(AODVMessageTypes.PROPOSE, self.componentinstancenumber,
                                    destNodeID,0,float('inf'),self.componentinstancenumber)
        payload = GenericMessagePayload(msgFromUser)
        message = GenericMessage(hdr, payload)
        self.send_down(Event(self,EventTypes.MFRT,message))
        self.MessageQueue[destNodeID] = msgFromUser
        self.lock.release()

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        self.MessageQueue = {}
        self.lock = threading.Lock()
        self.eventhandlers[AODVApplicationLayerEventType.PROPOSE] = self.on_propose

